/*
 * Copyright Contributors to the fusa-gcc-plugin project
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */
#include "macrohelpers.hpp"

using namespace std;

namespace macrohelpers {
// Get possible macro metadata for a specific location, loosely inspired by
// diagnostic.cc
optional<macro> get_macro_expansion(location_t where) {
  loc_map_pair loc;

  const line_map *map = linemap_lookup(line_table, where);

  if (!map)
    return nullopt;

  if (!linemap_macro_expansion_map_p(map))
    return nullopt;

  /**
   * Unwind macro expansions to get to the outermost macro
   *
   * 1| #define OP(x, op, y) (x op y)
   * 2| #define ADD(x, y) OP(x, +, y)
   * 3|
   * 4| int main() {
   * 5|    int result = ADD(1, 1);
   * 6| }
   *
   * In this case the expansion of macro at line (5) will
   * initially point to line (1) where `OP` definition lives,
   * however we want to report the macro name that was spelled
   * out in the source code, `ADD`, so we have to unwind the expansion
   * to find the macro definition at line (2).
   *
   * In other words, macro expansions are stored in the reverse
   * order that they are evaluated and appear in the source code.
   **/
  do {
    loc.where = where;
    loc.map = linemap_check_macro(map);

    where = linemap_unwind_toward_expansion(line_table, where, &map);
  } while (linemap_macro_expansion_map_p(map));

  location_t resolved_def_loc = linemap_resolve_location(
      line_table, loc.where, LRK_MACRO_DEFINITION_LOCATION, NULL);

  const line_map_ordinary *m = NULL;
  location_t l0 = linemap_resolve_location(line_table, resolved_def_loc,
                                           LRK_SPELLING_LOCATION, &m);

  if (IS_ADHOC_LOC(l0))
    l0 = get_location_from_adhoc_loc(line_table, l0);

  if (l0 < RESERVED_LOCATION_COUNT)
    return nullopt;

  /**
   * Detect if we're dealing with a complex macro, meaning:
   *  1) Functional macro with arguments, like `S_ISDIR(mode)`
   *  2) Macros with complex behaviours #define FOO (a + b)
   *
   * We distinguisg these with the following checks by looking
   * at the number of parameters (which should be zero), and the number
   * of tokens (which should be one). That way we don't spam
   * the user/system with this like errno which are trivial defines.
   *
   * There's, however, a class of macros that *may* fail this check
   * despite being trivial. An example of this is a macro like this:
   *
   *  #define FLAG (1 << 9)
   *
   * Trivial macros like this are handled by inspecting the tokens
   * and checking if there are any CPP_NAME tokens, if not then
   * the macro is comprised only of literals and operators.
   */
  const cpp_hashnode *node = loc.map->macro;
  if (node && node->type == NT_USER_MACRO) {
    const cpp_macro *macro = node->value.macro;
    size_t token_count = macro->count;

    bool has_name_ref = false;
    for (size_t i = 0; i < token_count; ++i) {
      const cpp_token *token = &macro->exp.tokens[i];

      has_name_ref = (token->type == CPP_NAME);
      if (has_name_ref)
        break;
    }

    if (!has_name_ref || !macro->fun_like ||
        (macro->paramc == 0 && token_count == 1))
      return nullopt;
  }

  return macro{linemap_map_get_macro_name(loc.map), LINEMAP_FILE(m), l0};
}
} // namespace macrohelpers
